cimport cython
from glypy.composition.ccomposition cimport CComposition
from glycopeptidepy._c.count_table cimport CountTable


cdef class ChemicalShiftBase(object):
    cdef:
        public str name
        public CComposition composition
        public double mass

    cpdef ChemicalShiftBase clone(self)


cdef class IonSeriesBase(object):
    cdef:
        public str name
        public int direction
        public bint includes_peptide
        public double mass_shift
        public CComposition composition_shift
        public Py_hash_t _hash
        public int int_code


cdef class FragmentBase(object):
    """Base class for all Fragment types. Defines basic
    name generation and neutral loss handling functions.

    Attributes
    ----------
    chemical_shift : ChemicalShift
        The ChemicalShift associated with this fragment, or None.
        If a ChemicalShift, its composition and mass are subtracted
        from this object's composition and mass attributes.
    name: str
        The human readable description of this fragment
    series: IonSeries
        The ion ladder this fragment is derived from

    """

    cdef:
        public str _name
        public Py_hash_t _hash
        public double mass
        public ChemicalShiftBase _chemical_shift

    cpdef IonSeriesBase get_series(self)

    cpdef ChemicalShiftBase get_chemical_shift(self)

    cpdef set_chemical_shift(self, ChemicalShiftBase chemical_shift)

    cpdef str base_name(self)
    cpdef str get_fragment_name(self)
    cdef void _update_hash_name(self)

    cdef str get_name(self)



cdef class SimpleFragment(FragmentBase):
    cdef:
        public IonSeriesBase kind
        public bint is_glycosylated
        public CComposition composition

    cpdef clone(self)


cdef class StubFragment(FragmentBase):
    cdef:
        public IonSeriesBase kind
        public bint is_glycosylated
        public CComposition composition
        public object glycosylation
        public bint is_extended
        int _glycosylation_size

    cpdef clone(self)

    @staticmethod
    cdef StubFragment _create(str name, double mass, IonSeriesBase kind, CComposition composition,
                              ChemicalShiftBase chemical_shift, bint is_glycosylated, object glycosylation,
                              bint is_extended, int glycosylation_size=*)

    cdef int get_glycosylation_size(self)


ctypedef fused mapping_types:
        CountTable
        dict
        object


cdef class PeptideFragment(FragmentBase):
    cdef:
        public IonSeriesBase kind
        public int position
        public CountTable modification_dict
        public double bare_mass
        public list flanking_amino_acids
        public dict glycosylation
        public CComposition composition

    @staticmethod
    cdef PeptideFragment _create(IonSeriesBase kind, int position, CountTable modification_dict, double mass,
                                 list flanking_amino_acids=*, dict glycosylation=*,
                                 ChemicalShiftBase chemical_shift=*, CComposition composition=?, double* delta_mass=?)

    cpdef clone(self)

    cdef void _update_mass_with_modifications(self)
    cdef bint _is_glycosylated(self)
    cdef long get_glycosylation_size(self) except -1


ctypedef fused list_or_iterable:
    list
    object


@cython.final
cdef class _NameTree(object):
    cdef:
        public object name
        public object value
        public dict children

    @staticmethod
    cdef _NameTree _create()

    cpdef _NameTree get(self, key)
    cpdef traverse(self, list_or_iterable parts)

cdef basestring build_name_from_composition(mapping_types glycan_composition)


ctypedef fused sequence_encoded_t:
    str
    object


cdef enum ParserState:
    start,
    n_term,
    aa,
    mod,
    c_term,


cdef object _cstring_sequence_tokenizer(str sequence, object implicit_n_term=*, object implicit_c_term=*, object glycan_parser_function=*)

cdef object _sequence_tokenizer(sequence_encoded_t sequence, object implicit_n_term=*, object implicit_c_term=*,
                                object glycan_parser_function=*)

cpdef sequence_tokenizer(object sequence, object implicit_n_term=*, object implicit_c_term=*,
                         object glycan_parser_function=*)

cpdef parse_simple(str sequence)


from typing import List
from glypy import GlycanComposition

from .. import constants as structure_constants
from ..glycan import (GlycosylationManager, glycosylation_site_detectors, GlycosylationType, GlycanCompositionWithOffsetProxy)
from ..modification import ModificationCategory, Modification
from ..residue import Residue
from ..fragmentation_strategy import OxoniumIonStrategy, StubGlycopeptideStrategy
from ..fragment import (
    _n_glycosylation, _o_glycosylation,
    _gag_linker_glycosylation)


@glycosylation_site_detectors(GlycosylationType.o_linked)
def find_o_glycosylation_sequons(sequence, allow_modified=frozenset()) -> List[int]:
    try:
        iter(allow_modified)
        allow_modified = set(allow_modified) | {Modification("HexNAc")}
    except TypeError:
        allow_modified = (Modification("HexNAc"),)

    positions = []
    ser = Residue("S")
    thr = Residue("T")
    asn = Residue("N")

    site_set = (ser, thr)

    for i, position in enumerate(sequence):
        if position[0] in site_set:
            if (not position.modifications or position.modifications[0] in allow_modified) and not (
                    i - 2 >= 0 and sequence[i - 2][0] == asn):
                positions.append(i)
    return positions


@glycosylation_site_detectors(GlycosylationType.n_linked)
def find_n_glycosylation_sequons(sequence, allow_modified=frozenset(), include_cysteine: bool = False) -> List[int]:
    try:
        iter(allow_modified)
        allow_modified = set(allow_modified) | {Modification(
            "HexNAc"), Modification(_o_glycosylation)}
    except TypeError:
        allow_modified = (Modification("HexNAc"),
                          Modification(_o_glycosylation))
    state = "seek"  # [seek, n, ^p, st]

    asn = Residue("Asn")
    pro = Residue("Pro")
    ser = Residue("Ser")
    thr = Residue("Thr")
    cys = Residue("Cys")

    if include_cysteine:
        p3 = {ser, thr, cys}
    else:
        p3 = {ser, thr}

    i = 0
    positions = []
    n_pos = None
    while(i < len(sequence)):
        next_pos = sequence[i]
        if state == "seek":
            # A sequon starts with an Asn residue without modifications, or for counting
            # purposes one that has already been glycosylated
            if next_pos.amino_acid == asn:
                if (not next_pos.modifications or next_pos.modifications[0] in allow_modified):
                    n_pos = i
                    state = "n"
        elif state == "n":
            if next_pos.amino_acid != pro:
                state = "^p"
            else:
                state = "seek"
                i = n_pos
                n_pos = None
        elif state == "^p":
            if next_pos.amino_acid in p3:
                positions.append(n_pos)
            i = n_pos
            n_pos = None
            state = "seek"
        i += 1
    return positions


@glycosylation_site_detectors(GlycosylationType.glycosaminoglycan)
def find_glycosaminoglycan_sequons(sequence, allow_modified=frozenset()) -> List[int]:
    try:
        iter(allow_modified)
        allow_modified = set(allow_modified) | {Modification(
            "Xyl"), Modification(_gag_linker_glycosylation)}
    except TypeError:
        allow_modified = (Modification("Xyl"),
                          Modification(_gag_linker_glycosylation))
    state = "seek"  # [seek, s, g1, x, g2]
    ser = Residue("Ser")
    # gly = Residue("Gly")

    i = 0
    positions = []

    # s_position = None
    while(i < len(sequence)):
        next_pos = sequence[i]
        if state == "seek":
            if next_pos.amino_acid == ser:
                if (not next_pos.modifications or next_pos.modifications[0] in allow_modified):
                    positions.append(i)
        i += 1
    return positions


class GlycosylatedSequenceMixin(object):
    __slots__ = ()

    @property
    def glycan(self):
        return self.glycan_composition

    @glycan.setter
    def glycan(self, value):
        if isinstance(value, (GlycanComposition, GlycanCompositionWithOffsetProxy)):
            self._glycosylation_manager.aggregate = value
        elif isinstance(value, GlycosylationManager):
            self._glycosylation_manager = value # pylint: disable=assigning-non-slot
        elif value is None:
            pass
        else:
            raise TypeError("Cannot convert %s %r" % (type(value), value))
        self._invalidate()

    @property
    def glycosylation_manager(self):
        return self._glycosylation_manager

    @property
    def glycan_composition(self):
        return self._glycosylation_manager.glycan_composition

    @property
    def total_glycosylation_size(self):
        return self._glycosylation_manager.total_glycosylation_size()

    @property
    def n_glycan_sequon_sites(self):
        return find_n_glycosylation_sequons(self, structure_constants.ALLOW_MODIFIED_ASPARAGINE)

    @property
    def o_glycan_sequon_sites(self):
        return find_o_glycosylation_sequons(self)

    @property
    def glycosaminoglycan_sequon_sites(self):
        return find_glycosaminoglycan_sequons(self)

    @property
    def glycosylation_sites(self):
        return self.n_glycan_sequon_sites

    def deglycosylate(self):
        self._invalidate()
        _glycosylation_enum = ModificationCategory.glycosylation
        for i, pos in enumerate(self):
            if pos.modifications:
                mods = [mod.name for mod in pos.modifications if mod.is_a(
                    _glycosylation_enum)]
                for mod in mods:
                    self.drop_modification(i, mod)
        self._glycosylation_manager.clear()
        # This line is a no-op.
        # self.glycan = None
        return self

    def glycan_fragments(self, oxonium=True, all_series=False, allow_ambiguous=False,
                         include_large_glycan_fragments=True, maximum_fragment_size=5,
                         strategy=None):
        r'''
        Generate all oxonium ions for the attached glycan, and
        if `all_series` is `True`, then include the B/Y glycan
        ion ladder, with the peptide attached to the Y ion ladder.

        Parameters
        ----------
        all_series: bool
            Generate the B/Y+peptide ion ladder, otherwise just 2-residue
            pairs for all monosaccharides in :attr:`self.glycan`

        Yields
        ------
        SimpleFragment
        '''
        if strategy is None:
            strategy = OxoniumIonStrategy
        return strategy(
            self, oxonium=oxonium, all_series=all_series, allow_ambiguous=allow_ambiguous,
            include_large_glycan_fragments=include_large_glycan_fragments,
            maximum_fragment_size=maximum_fragment_size)

    def stub_fragments(self, extended=False, extended_fucosylation=False, strategy=None, **kwargs):
        if strategy is None:
            strategy = StubGlycopeptideStrategy
        return strategy(
            self, extended, extended_fucosylation=extended_fucosylation, **kwargs)

    def is_multiply_glycosylated(self):
        return len(self.glycosylation_manager) > 1

    def is_o_glycosylated(self):
        return self.glycosylation_manager.count_glycosylation_type(GlycosylationType.o_linked)

    def is_n_glycosylated(self):
        return self.glycosylation_manager.count_glycosylation_type(GlycosylationType.n_linked)

    def is_gag_linker_glycosylated(self):
        return self.glycosylation_manager.count_glycosylation_type(GlycosylationType.glycosaminoglycan)



import logging
import sys
from . import constants as structure_constants
from . import PeptideSequenceBase
from .modification import NGlycanCoreGlycosylation


from glypy.structure.glycan_composition import FrozenGlycanComposition

_n_glycan_core = NGlycanCoreGlycosylation()
glycan_parser = FrozenGlycanComposition.parse


def sequence_tokenizer(sequence, implicit_n_term=None, implicit_c_term=None, glycan_parser_function=None):
    '''A simple stateful sequence parser implementing a formally context-free language
    describing components of a polypeptide sequence with N-, C- and internal modifications,
    as well as a glycan composition written at the end.

    Return
    ------
    chunks: list
    mods: list
    n_term: str
    c_term: str
    glycan: :class:`glypy.composition.glycan_composition.FrozenGlycanComposition`
    '''
    if glycan_parser_function is None:
        glycan_parser_function = glycan_parser
    if sys.version_info[0] > 2 and isinstance(sequence, bytes):
        sequence = sequence.decode("utf8")

    state = "start"  # [start, n_term, aa, mod, c_term]
    n_term = implicit_n_term or structure_constants.N_TERM_DEFAULT
    c_term = implicit_c_term or structure_constants.C_TERM_DEFAULT
    mods = []
    chunks = []
    glycan = ""
    current_aa = ""
    current_mod = ""
    current_mods = []
    paren_level = 0
    i = 0
    while i < len(sequence):
        next_char = sequence[i]

        # Transition from aa to mod when encountering the start of a modification
        # internal to the sequence
        if next_char == "(":
            if state == "aa":
                state = "mod"
                assert paren_level == 0
                paren_level += 1

            # Transition to n_term when starting on an open parenthesis
            elif state == "start":
                state = "n_term"
                paren_level += 1

            else:
                paren_level += 1
                if not (state in {"n_term", "c_term"} and paren_level == 1):
                    current_mod += next_char

        elif next_char == ")":
            if state == "aa":
                raise Exception(
                    "Invalid Sequence. ) found outside of modification, Position {0}. {1}".format(i, sequence))
            else:
                paren_level -= 1
                if paren_level == 0:
                    mods.append(current_mod)
                    current_mods.append(current_mod)
                    if state == "mod":
                        state = 'aa'
                        # If we encounter multiple modifications in separate parentheses
                        # in a row, attach them to the previous position
                        if current_aa == "":
                            chunks[-1][1].extend(current_mods)
                        else:
                            chunks.append([current_aa, current_mods])

                    elif state == "n_term":
                        if sequence[i + 1] != "-":
                            raise Exception("Malformed N-terminus for " + sequence)
                        # Only one modification on termini
                        n_term = current_mod
                        state = "aa"
                        # Jump ahead past - into the amino acid sequence
                        i += 1

                    elif state == "c_term":
                        # Only one modification on termini
                        c_term = current_mod

                    current_mods = []
                    current_mod = ""
                    current_aa = ""
                else:
                    current_mod += next_char

        elif next_char == "|":
            if state == "aa":
                raise Exception("Invalid Sequence. | found outside of modification")
            else:
                current_mods.append(current_mod)
                mods.append(current_mod)
                current_mod = ""

        elif next_char == "{":
            if (state == 'aa' or (state == "c_term" and paren_level == 0)):
                glycan = sequence[i:]
                break
            elif (state in {"mod", "n_term", "c_term"}) and paren_level > 0:
                current_mod += next_char

        elif next_char == "-":
            if state == "aa":
                state = "c_term"
                if(current_aa != ""):
                    current_mods.append(current_mod)
                    chunks.append([current_aa, current_mods])
                    current_mod = ""
                    current_mods = []
                    current_aa = ""
            else:
                current_mod += next_char

        elif state == "start":
            state = "aa"
            current_aa = next_char
        elif state == "aa":
            if(current_aa != ""):
                current_mods.append(current_mod)
                chunks.append([current_aa, current_mods])
                current_mod = ""
                current_mods = []
                current_aa = ""
            current_aa = next_char
        elif state in {"n_term", "mod", "c_term"}:
            current_mod += next_char
        else:
            raise Exception(
                "Unknown Tokenizer State", current_aa, current_mod, i, next_char)
        i += 1
    if current_aa != "":
        current_mods.append("")
        chunks.append([current_aa, current_mods])
    if current_mod != "":
        mods.append(current_mod)

    if glycan != "":
        try:
            glycan = glycan_parser_function(glycan)
        except Exception as e:
            logging.exception("Error in parser, %s and %s", glycan, sequence, exc_info=e)
    else:
        glycan = None

    return chunks, mods, glycan, n_term, c_term


try:
    _sequence_tokenizer = sequence_tokenizer
    from glycopeptidepy._c.parser import sequence_tokenizer
except ImportError:
    pass


def prefix_to_postfix_modifications(sequence):
    '''Transform a chunk-list from modification-on-the-left to
    modification-on-the-right association rules.

    This code does not work with terminal modifications

    Return
    ------
    list
    '''
    sequence = "!" + sequence

    tokens, mods, glycan, n_term, c_term = sequence_tokenizer(sequence)
    last_mods = tokens.pop(0)[1]
    for token in tokens:
        mods = token[1]
        token[1] = last_mods
        last_mods = mods
    return tokens


def sequence_length(sequence):
    '''Compute the peptide length of a sequence string.
    This parses the sequence and counts the number of chunks.

    Return
    ------
    int
    '''
    sequence, modifications, glycan, n_term, c_term = sequence_tokenizer(sequence)
    return len(sequence)


def strip_modifications(sequence):
    '''A wrapper around `sequence_tokenizer` that discards all modifications

    Return
    ------
    str
    '''
    if isinstance(sequence, PeptideSequenceBase):
        sequence = str(sequence)
    chunks, _, glycan, n_term, c_term = sequence_tokenizer(sequence)
    return ''.join([residue for residue, mod in chunks])


def sequence_tokenizer_respect_sequons(sequence, known_sites=None):
    '''A wrapper around `sequence_tokenizer` that will treat an n-glycan sequon as a single unit

    Return
    ------
    list
    '''
    if known_sites is None:
        known_sites = tuple()
    chunks, _, glycan, n_term, c_term = sequence_tokenizer(sequence)
    positions = []
    i = 0
    sequence_length = len(chunks)
    while(i < sequence_length):
        cur_pos = chunks[i]
        if (cur_pos[0] == "N" and cur_pos[1] and _n_glycan_core in cur_pos[1]) or (i in known_sites):
            positions.append(chunks[i:(i + 3)])
            i += 2
        else:
            positions.append(cur_pos)
        i += 1
    return positions


def parse_simple(sequence):
    '''Parse a simple sequence string without *any* modifications
    into a chunked token list, matching the return signature of :func:`sequence_tokenizer`
    '''
    if sys.version_info[0] > 2 and isinstance(sequence, bytes):
        sequence = sequence.decode("utf8")

    chunks = [[a, []] for a in sequence]
    mods = None
    glycan = None
    n_term = structure_constants.N_TERM_DEFAULT
    c_term = structure_constants.C_TERM_DEFAULT
    return chunks, mods, glycan, n_term, c_term


try:
    from glycopeptidepy._c.parser import parse_simple
except ImportError:
    pass



import sys
import os
import re
import traceback

from setuptools import setup, find_packages, Extension as _Extension

from distutils.command.build_ext import build_ext
from distutils.errors import (CCompilerError, DistutilsExecError,
                              DistutilsPlatformError)


def has_option(name):
    try:
        sys.argv.remove('--%s' % name)
        return True
    except ValueError:
        pass
    # allow passing all cmd line options also as environment variables
    env_val = os.getenv(name.upper().replace('-', '_'), 'false').lower()
    if env_val == "true":
        return True
    return False


include_diagnostics = has_option("include-diagnostics")
force_cythonize = has_option("force-cythonize")


def Extension(*args, **kwargs):
    ext = _Extension(*args, **kwargs)
    ext.include_dirs.append("src/glycopeptidepy/_c")
    return ext


def make_extensions():
    is_ci = bool(os.getenv("CI", ""))
    macros = []
    try:
        from Cython.Build import cythonize
        cython_directives = {
            'embedsignature': True,
            "profile": include_diagnostics
        }
        if include_diagnostics:
            macros.append(("CYTHON_TRACE_NOGIL", "1"))
        if is_ci and include_diagnostics:
            cython_directives['linetrace'] = True
        extensions = cythonize([
            Extension("glycopeptidepy._c.collectiontools", sources=['src/glycopeptidepy/_c/collectiontools.pyx']),
            Extension("glycopeptidepy._c.count_table", sources=[
                      'src/glycopeptidepy/_c/count_table.pyx']),
            Extension("glycopeptidepy._c.parser", sources=[
                      'src/glycopeptidepy/_c/parser.pyx']),
            Extension("glycopeptidepy._c.structure.base", sources=[
                      'src/glycopeptidepy/_c/structure/base.pyx']),
            Extension("glycopeptidepy._c.structure.fragment",
                      sources=['src/glycopeptidepy/_c/structure/fragment.pyx']
                     ),
            Extension("glycopeptidepy._c.structure.constants", sources=[
                      'src/glycopeptidepy/_c/structure/constants.pyx']),
            Extension("glycopeptidepy._c.structure.glycan", sources=[
                      'src/glycopeptidepy/_c/structure/glycan.pyx']),
            Extension("glycopeptidepy._c.structure.sequence_methods",
                      sources=['src/glycopeptidepy/_c/structure/sequence_methods.pyx']),
            Extension("glycopeptidepy._c.structure.modification.rule",
                      sources=['src/glycopeptidepy/_c/structure/modification/rule.pyx']),
            Extension("glycopeptidepy._c.structure.modification.modification",
                      sources=['src/glycopeptidepy/_c/structure/modification/modification.pyx']),
            Extension("glycopeptidepy._c.structure.modification.source",
                      sources=['src/glycopeptidepy/_c/structure/modification/source.pyx']),
            Extension("glycopeptidepy._c.structure.fragmentation_strategy.base",
                      sources=["src/glycopeptidepy/_c/structure/fragmentation_strategy/base.pyx"]),
            Extension("glycopeptidepy._c.structure.fragmentation_strategy.peptide",
                      sources=["src/glycopeptidepy/_c/structure/fragmentation_strategy/peptide.pyx"]),
            Extension("glycopeptidepy._c.structure.fragmentation_strategy.glycan",
                      sources=["src/glycopeptidepy/_c/structure/fragmentation_strategy/glycan.pyx"]),
            Extension("glycopeptidepy._c.algorithm",
                      sources=['src/glycopeptidepy/_c/algorithm.pyx']),
        ], compiler_directives=cython_directives, force=force_cythonize)
    except ImportError:
        extensions = [
            Extension("glycopeptidepy._c.collectiontools", sources=['src/glycopeptidepy/_c/collectiontools.c']),
            Extension("glycopeptidepy._c.count_table", sources=['src/glycopeptidepy/_c/count_table.c']),
            Extension("glycopeptidepy._c.parser", sources=['src/glycopeptidepy/_c/parser.c']),
            Extension("glycopeptidepy._c.structure.base", sources=['src/glycopeptidepy/_c/structure/base.c']),
            Extension("glycopeptidepy._c.structure.fragment", sources=['src/glycopeptidepy/_c/structure/fragment.c']),
            Extension("glycopeptidepy._c.structure.constants", sources=['src/glycopeptidepy/_c/structure/constants.c']),
            Extension("glycopeptidepy._c.structure.glycan", sources=[
                      'src/glycopeptidepy/_c/structure/glycan.c']),
            Extension("glycopeptidepy._c.structure.sequence_methods",
                      sources=['src/glycopeptidepy/_c/structure/sequence_methods.c']),
            Extension("glycopeptidepy._c.structure.modification.rule",
                      sources=['src/glycopeptidepy/_c/structure/modification/rule.c']),
            Extension("glycopeptidepy._c.structure.modification.modification",
                      sources=['src/glycopeptidepy/_c/structure/modification/modification.c']),
            Extension("glycopeptidepy._c.structure.modification.source",
                      sources=['src/glycopeptidepy/_c/structure/modification/source.c']),
            Extension("glycopeptidepy._c.structure.fragmentation_strategy.base",
                      sources=["src/glycopeptidepy/_c/structure/fragmentation_strategy/base.c"]),
            Extension("glycopeptidepy._c.structure.fragmentation_strategy.peptide",
                      sources=["src/glycopeptidepy/_c/structure/fragmentation_strategy/peptide.c"]),
            Extension("glycopeptidepy._c.structure.fragmentation_strategy.glycan",
                      sources=["src/glycopeptidepy/_c/structure/fragmentation_strategy/glycan.c"]),
            Extension("glycopeptidepy._c.algorithm", sources=['src/glycopeptidepy/_c/algorithm.c']),
        ]
    return extensions


ext_errors = (CCompilerError, DistutilsExecError, DistutilsPlatformError)
if sys.platform == 'win32':
    # 2.6's distutils.msvc9compiler can raise an IOError when failing to
    # find the compiler
    ext_errors += (IOError,)


class BuildFailed(Exception):

    def __init__(self):
        self.cause = sys.exc_info()[1]  # work around py 2/3 different syntax

    def __str__(self):
        return str(self.cause)


class ve_build_ext(build_ext):
    # This class allows C extension building to fail.

    def run(self):
        try:
            build_ext.run(self)
        except DistutilsPlatformError:
            traceback.print_exc()
            raise BuildFailed()

    def build_extension(self, ext):
        try:
            build_ext.build_extension(self, ext)
        except ext_errors:
            traceback.print_exc()
            raise BuildFailed()
        except ValueError:
            # this can happen on Windows 64 bit, see Python issue 7511
            traceback.print_exc()
            if "'path'" in str(sys.exc_info()[1]):  # works with both py 2/3
                raise BuildFailed()
            raise


cmdclass = {}

cmdclass['build_ext'] = ve_build_ext


def status_msgs(*msgs):
    print('*' * 75)
    for msg in msgs:
        print(msg)
    print('*' * 75)


def version():
    return re.sub(
        r'[\s\'"\n]',
        '',
        open("src/glycopeptidepy/version.py").readline().split("=")[1]
    )


required = []
with open('requirements.txt') as f:
    required = f.read().splitlines()


def run_setup(include_cext=True):
    setup(
        name='glycopeptidepy',
        version=version(),
        packages=find_packages(where='src', include='glycopeptidepy*'),
        install_requires=required,
        include_package_data=True,
        package_dir={"": "src"},
        ext_modules=make_extensions() if include_cext else None,
        cmdclass=cmdclass,
        zip_safe=False,
        package_data={
            'src/glycopeptidepy': ["*.csv", "*.xml", "*.json", "data/*.csv"],
            'src/glycopeptidepy/structure': [
                "structure/modification/data/*.csv",
                "structure/modification/data/*.json"]
        },
    )


run_setup(True)
# try:
#     run_setup(True)
# except Exception as exc:
#     run_setup(False)

#     status_msgs(
#         "WARNING: The C extension could not be compiled, " +
#         "speedups are not enabled.",
#         "Plain-Python build succeeded."
#     )



# pylint: disable=assigning-non-slot
from ..base import SequencePosition
from ..modification import ModificationCategory, Modification, SequenceLocation
from ..terminal_group import _make_terminal_group
from .. import constants as structure_constants


class MutableSequenceMixin(object):
    __slots__ = ()

    def drop_modification(self, position, modification_type):
        '''
        Drop a modification by name from a specific residue. If the
        position is the N-term or the C-term, the terminal modification will
        be reset to the default.

        Parameters
        ----------
        position: int
            The position of the modification to drop
        modification_type: str or Modification
            The modification to drop

        Raises
        ------
        ValueError:
            If the `modification_type` is not found at `position`, a :class:`ValueError` will
            be raised.
        '''
        dropped_index = None
        self._invalidate()
        if position is SequenceLocation.n_term:
            self.n_term = _make_terminal_group(structure_constants.N_TERM_DEFAULT)
            return
        elif position is SequenceLocation.c_term:
            self.c_term = _make_terminal_group(structure_constants.C_TERM_DEFAULT)
            return

        for i, mod in enumerate(self.sequence[position].modifications):
            if modification_type == mod.rule:
                dropped_index = i
                break
        if dropped_index is None:
            raise ValueError("Modification not found! %s @ %s" %
                             (modification_type, position))
        try:
            drop_mod = self.sequence[position].modifications.pop(dropped_index)
            self.mass -= drop_mod.mass
            if drop_mod.is_tracked_for(ModificationCategory.glycosylation):
                self._glycosylation_manager.pop(position)
        except (IndexError, ValueError):
            raise ValueError("Modification not found! %s @ %s" %
                             (modification_type, position))

    def add_modification(self, position, modification_type):
        '''
        Add a modification by name to a specific residue. If the
        position is the N-term or the C-term, the terminal modification will
        be replaced.

        Parameters
        ----------
        position: int
            The position of the modification to add
        modification_type: str or Modification
            The modification to add
        '''
        self._invalidate()
        if isinstance(modification_type, Modification):
            mod = modification_type
        else:
            mod = Modification(rule=modification_type)

        if position is SequenceLocation.n_term:
            self.n_term = self.n_term.modify(mod)
        elif position is SequenceLocation.c_term:
            self.c_term = self.c_term.modify(mod)
        else:
            if (position == -1) or (position >= len(self.sequence)):
                raise IndexError(
                    "Invalid modification position. %s, %s, %s" %
                    (position, str(self.sequence), modification_type))

            self.sequence[position][1].append(mod)
            self.mass += mod.mass
            if mod.is_tracked_for(ModificationCategory.glycosylation):
                self._glycosylation_manager[position] = mod

    def _retrack_sequence(self):
        self._glycosylation_manager.clear()
        for i, position in enumerate(self):
            if position.modifications:
                for mod in position.modifications:
                    if mod.is_tracked_for(ModificationCategory.glycosylation):
                        self._glycosylation_manager[i] = mod

    def insert(self, position, residue, modifications=None):
        if modifications is None:
            modifications = []
        self._invalidate()
        self.sequence.insert(
            position, SequencePosition([residue, modifications]))
        self.mass += residue.mass
        for mod in modifications:
            self.mass += mod.mass
        self._retrack_sequence()

    def delete(self, position):
        self._invalidate()
        residue, mods = self.sequence.pop(position)
        self.mass -= residue.mass
        if mods:
            for mod in mods:
                self.mass -= mod.mass
        self._retrack_sequence()

    def substitute(self, position, residue):
        old_residue = self.sequence[position].amino_acid
        self.mass -= old_residue.mass
        self.mass += residue.mass
        self.sequence[position].amino_acid = residue
        self._invalidate()
        self._retrack_sequence()

    def append(self, residue, modification=None):
        self._invalidate()
        self.mass += residue.mass
        next_pos = [residue]
        if modification is None:
            next_pos.append([])
        else:
            next_pos.append([modification])
            self.mass += modification.mass
        self.sequence.append(SequencePosition(next_pos))
        self._retrack_sequence()

    def extend(self, sequence):
        self._invalidate()
        self.sequence.extend(sequence.sequence)
        self.mass += sequence.mass - sequence.n_term.mass - sequence.c_term.mass
        self._retrack_sequence()

    def leading_extend(self, sequence):
        self._invalidate()
        self.sequence = sequence.sequence + self.sequence
        self.mass += sequence.mass - sequence.n_term.mass - sequence.c_term.mass
        self._retrack_sequence()


try:
    from glycopeptidepy._c.structure.sequence_methods import add_modification as _c_add_modification

    MutableSequenceMixin.add_modification = _c_add_modification
except ImportError:
    pass



'''Represents fragments of peptides and glycopeptides.
'''
import re
from collections import defaultdict
from six import add_metaclass
from .modification import (
    Modification, NGlycanCoreGlycosylation, OGlycanCoreGlycosylation,
    GlycosaminoglycanLinkerGlycosylation, ModificationCategory)
from .glycan import HashableGlycanComposition
from .composition import Composition, formula

_n_glycosylation = NGlycanCoreGlycosylation()
_o_glycosylation = OGlycanCoreGlycosylation()
_gag_linker_glycosylation = GlycosaminoglycanLinkerGlycosylation()
_modification_hexnac = Modification("HexNAc").rule
_modification_xylose = Modification("Xyl").rule

fragment_pairing = {
    "a": "x",
    "b": "y",
    "c": "z",
    "x": "a",
    "y": "b",
    "z": "c",
}

fragment_shift_composition = {
    'a': -Composition("CHO"),
    'b': -Composition({"H": 1}),
    'c': Composition("NH2"),
    'x': Composition("CO2") - Composition("H"),
    'y': Composition('H'),
    'z': (-Composition("NH")),
    'zp': (-Composition("NH2")),
}

fragment_shift = {
    k: v.mass for k, v in fragment_shift_composition.items()
}


fragment_direction = {
    "a": 1,
    "b": 1,
    "c": 1,
    "x": -1,
    "y": -1,
    "z": -1,
    "zp": -1,
}

generic_chemical_shifts_composition = {
    "-NH3": -Composition("NH3"),
    "-H2O": -Composition("H2O"),
    # "-NH3-NH3": -Composition("(NH3)2"),
    "-H2O-H2O": -Composition("(H2O)2"),
    "-NH3-H2O": -Composition("NH3H2O")
}


def format_negative_composition(composition):
    '''Format a chemical composition as a formula with negative
    element counts.
    '''
    return "-%s" % formula({k: -v for k, v in composition.items()})

# pylint: disable=access-member-before-definition,assigning-non-slot,invalid-unary-operand-type


try:
    from glycopeptidepy._c.structure.fragment import ChemicalShiftBase as _ChemicalShiftBase
except ImportError:
    _ChemicalShiftBase = object


class ChemicalShift(_ChemicalShiftBase):
    '''Represents a chemical shift, a gain or loss of chemical composition.

    Attributes
    ----------
    name: str
        The name of the modification
    composition: Composition
        The chemical composition difference
    mass: float
        The mass of the chemical composition shift
    '''
    def __init__(self, name, composition=None):
        if composition is None:
            composition = generic_chemical_shifts_composition[name]
        self.name = name
        self.composition = composition
        self.mass = composition.mass

    def clone(self):
        '''Create a copy of this object

        Returns
        -------
        :class:`ChemicalShift`
        '''
        return self.__class__(self.name, self.composition.clone())

    def __str__(self):
        return self.name

    def __repr__(self):
        return "%s(name=%r)" % (self.__class__.__name__, self.name)

    def is_loss(self):
        '''Returns :const:`True` if the chemical shift results in a net loss in mass

        Returns
        -------
        :class:`bool`
        '''
        return self.mass < 0

    AllLosses = object()


NeutralLoss = ChemicalShift


class FragmentBase(object):
    """Base class for all Fragment types. Defines basic
    name generation and neutral loss handling functions.

    Attributes
    ----------
    chemical_shift : ChemicalShift
        The ChemicalShift associated with this fragment, or None.
        If a ChemicalShift, its composition and mass are subtracted
        from this object's composition and mass attributes.
    name: str
        The human readable description of this fragment
    series: IonSeries
        The ion ladder this fragment is derived from

    """

    __slots__ = ("_chemical_shift", "_name", "_hash")

    def get_series(self):
        '''Return the :class:`IonSeries` this fragment belongs to.

        Returns
        -------
        :class:`IonSeries`
        '''
        raise NotImplementedError()

    def __hash__(self):
        if self._hash is None:
            self._hash = hash(self.name)
        return self._hash

    def __eq__(self, other):
        try:
            return self.name == other.name and abs(self.mass - other.mass) < 1e-5
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self == other

    @property
    def series(self):
        '''The :class:`IonSeries` this fragment is drawn from.

        Returns
        -------
        :class:`IonSeries`
        '''
        return self.get_series()

    def clone(self):
        '''Create a copy of this object

        Returns
        -------
        :class:`FragmentBase`
        '''
        raise NotImplementedError()

    def get_chemical_shift(self):
        '''Returns the chemical shift associated with the fragment.

        Returns
        -------
        :class:`ChemicalShift`
        '''
        return self._chemical_shift

    def set_chemical_shift(self, chemical_shift):
        '''Sets the chemical shift associated with the fragment, updating
        the :attr:`mass` and :attr:`chemical_shift`.
        '''
        if self._chemical_shift is not None:
            self.mass -= self._chemical_shift.mass
        self._chemical_shift = chemical_shift
        if chemical_shift is not None:
            self.mass += chemical_shift.mass

    def get_fragment_name(self):
        parts = [self._name]
        chemical_shift = self.chemical_shift
        if chemical_shift is not None:
            parts.append(str(chemical_shift))
        return ''.join(parts)

    def base_name(self):
        return self._name

    @property
    def name(self):
        if self._name is None:
            self._name = self.get_fragment_name()
        return self._name

    @name.setter
    def set_name(self, name):
        self._name = name

    chemical_shift = property(get_chemical_shift, set_chemical_shift)

    def generate_chemical_shifts(self, losses=ChemicalShift.AllLosses):
        if losses is ChemicalShift.AllLosses:
            losses = generic_chemical_shifts_composition.keys()

        for loss in losses:
            frag = self.clone()
            if loss is not None:
                frag.chemical_shift = ChemicalShift(loss)
            yield frag


class PeptideFragment(FragmentBase):
    '''Represents a peptide backbone fragment, such as a peptide b or y ion.

    Attributes
    ----------
    series: :class:`IonSeries`
        The ion series this fragment is from
    bare_mass: float
        The unmodified mass of this fragment
    mass: float
        The modified, correct mass of this fragment
    position: int
        The peptide backbone position index this fragment came from, relative to the
        starting terminal
    modification_dict: :class:`~.ModificationIndex`
        A count of the modifications on this fragment
    flanking_amino_acids: list
        The adjacent amino acids of the amide bond that broke to form this fragment
    glycosylation: dict
        A position to glycan mapping for glycan structures
    chemical_shift: :class:`ChemicalShift`
        An associated chemical shift
    composition: :class:`~.Composition`
        The elemental composition of this fragment
    '''
    concerned_modifications = set(
        [_n_glycosylation,
         _modification_hexnac,
         _o_glycosylation,
         _gag_linker_glycosylation,
         _modification_xylose])

    __slots__ = ("kind", "position", "modification_dict", "bare_mass",
                 "flanking_amino_acids", "glycosylation",
                 "_chemical_shift", "composition", 'mass')

    def __init__(self, kind, position, modification_dict, mass,
                 flanking_amino_acids=None, glycosylation=None, chemical_shift=None,
                 composition=None):
        self.kind = kind

        # The mass value is the bare backbone's mass
        self.bare_mass = mass
        self.modification_dict = modification_dict
        self.mass = mass
        self.composition = composition
        self._chemical_shift = None
        self._name = None
        self._hash = None

        self.flanking_amino_acids = flanking_amino_acids
        self.position = position

        self.glycosylation = glycosylation
        self.chemical_shift = chemical_shift

        self._update_mass_with_modifications()

    def _update_mass_with_modifications(self):
        for key, value in self.modification_dict.items():
            self.mass += (key).mass * value
        chemical_shift = self.get_chemical_shift()
        if chemical_shift is not None:
            self.mass += chemical_shift.mass

    def get_series(self):
        return self.kind

    def clone(self):
        return self.__class__(
            self.series, self.position, dict(self.modification_dict),
            self.bare_mass, tuple(
                self.flanking_amino_acids),
            self.glycosylation.clone() if self.glycosylation is not None else None,
            self._chemical_shift.clone() if self._chemical_shift is not None else None,
            self.composition.clone())

    def total_composition(self):
        composition = self.composition.clone()
        chemical_shift = self.chemical_shift
        if chemical_shift is not None:
            composition += chemical_shift.composition
        return composition

    def __reduce__(self):
        return self.__class__, (
            self.kind, self.position, self.modification_dict, self.bare_mass,
            self.flanking_amino_acids, self.glycosylation,
            self.chemical_shift, self.composition)

    def base_name(self):
        """Simply return string like b2, y3 with no modification information."""
        fragment_name = []
        fragment_name.append(str(self.series))
        fragment_name.append(str(self.position))
        return ''.join(fragment_name)

    def get_fragment_name(self):
        """Build a complete fragment name listing any glycosylation modifications
        """
        fragment_name = []
        fragment_name.append(str(self.series))
        fragment_name.append(str(self.position))

        # Only concerned modifications are reported.
        for mod_rule, count in self.modification_dict.items():
            if mod_rule in self.concerned_modifications or mod_rule.is_a(ModificationCategory.glycosylation):
                if count > 1:
                    fragment_name.extend(
                        ['+', str(count), str(mod_rule.name)])
                elif count == 1:
                    fragment_name.extend(['+', str(mod_rule.name)])
                else:
                    pass

        if self.chemical_shift is not None:
            fragment_name.append(str(self.chemical_shift))
        name = ''.join(fragment_name)
        return name

    @property
    def is_glycosylated(self):
        if self.glycosylation:
            return True
        else:
            for mod in self.modification_dict:
                if mod in self.concerned_modifications or mod.is_a(ModificationCategory.glycosylation):
                    return True
        return False

    @property
    def glycosylation_size(self):
        if self.glycosylation is not None:
            raise NotImplementedError()
        size = 0
        size += self.modification_dict.get(_modification_hexnac, 0)
        size += self.modification_dict.get(_modification_xylose, 0)
        return size

    def __repr__(self):
        template = ("PeptideFragment({self.series}, {self.position}, "
                    "{self.mass}, {modification_dict}, {self.flanking_amino_acids},"
                    " {self.chemical_shift})")
        return template.format(self=self, modification_dict=dict(self.modification_dict))



try:
    _PeptideFragment = PeptideFragment
    from glycopeptidepy._c.structure.fragment import PeptideFragment as _CPeptideFragment
    PeptideFragment = _CPeptideFragment
except ImportError:
    pass


class SimpleFragment(FragmentBase):
    __slots__ = ["name", "mass", "kind", "composition",
                 "_chemical_shift", "is_glycosylated"]

    def __init__(self, name, mass, kind, composition, chemical_shift=None, is_glycosylated=False):
        self._name = None
        self._hash = None
        self._chemical_shift = None
        self.name = name
        self.mass = mass
        self.kind = kind
        self.chemical_shift = chemical_shift
        self.composition = composition
        self.is_glycosylated = is_glycosylated

    def clone(self):
        corrected_mass = self.mass
        if self._chemical_shift is not None:
            corrected_mass -= self.chemical_shift.mass
        return self.__class__(self.name, corrected_mass, self.kind,
                              self.composition,
                              self._chemical_shift.clone() if self._chemical_shift is not None else None,
                              self.is_glycosylated)

    def __reduce__(self):
        corrected_mass = self.mass
        if self._chemical_shift is not None:
            corrected_mass -= self.chemical_shift.mass
        return self.__class__, (self.name, corrected_mass, self.kind, self.composition,
                                self.chemical_shift, self.is_glycosylated)

    def __repr__(self):
        return ("{self.__class__.__name__}(name={self.name}, "
                "mass={self.mass:.04f}, series={self.kind})").format(self=self)

    def get_series(self):
        return self.kind


try:
    _SimpleFragment = SimpleFragment
    from glycopeptidepy._c.structure.fragment import SimpleFragment as _CSimpleFragment
    SimpleFragment = _CSimpleFragment
except ImportError:
    pass


try:
    from glycopeptidepy._c.structure.fragment import _NameTree
except ImportError:

    class _NameTree(object):
        def __init__(self, name=None, children=None):
            if children is None:
                children = defaultdict(_NameTree)
            self.name = name
            self.children = children

        def __getitem__(self, key):
            return self.children[key]

        def traverse(self, parts):
            node = self
            for kv in parts:
                node = node[kv]
            return node


class StubFragment(SimpleFragment):
    __slots__ = ['glycosylation', 'is_extended']

    _name_cache = _NameTree()

    def __init__(self, name, mass, kind, composition, chemical_shift=None, is_glycosylated=False,
                 glycosylation=None, is_extended=False):
        if glycosylation is None:
            glycosylation = HashableGlycanComposition()
        super(StubFragment, self).__init__(name, mass, kind, composition, chemical_shift, is_glycosylated)
        self.glycosylation = glycosylation
        self.is_extended = is_extended

    def clone(self):
        dup = super(StubFragment, self).clone()
        dup.glycosylation = self.glycosylation
        dup.is_extended = self.is_extended
        return dup

    @property
    def glycosylation_size(self):
        return sum(self.glycosylation.values())

    @classmethod
    def build_name_from_composition(cls, glycan_composition):
        name = 'peptide'
        root = cls._name_cache
        parts = list(glycan_composition.items())
        # traverse the tree to find the node referring to
        # this combination of monosaccharides and counts
        node = root.traverse(parts)
        extended_key = node.name
        if extended_key is None:
            parts.sort(key=lambda x: x[0].mass())
            node.name = extended_key = ''.join("%s%d" % kv for kv in parts)
        if extended_key:
            name = "%s+%s" % (name, extended_key)
        return name

    def __reduce__(self):
        proto = list(super(StubFragment, self).__reduce__())
        proto[1] = proto[1] + (self.glycosylation, self.is_extended)
        return tuple(proto)

    def base_name(self):
        if self.chemical_shift is None:
            return self._name
        else:
            return self.build_name_from_composition(self.glycosylation)


try:
    from glycopeptidepy._c.structure.fragment import StubFragment
except ImportError:
    pass


class MemoizedIonSeriesMetaclass(type):

    def __call__(cls, name=None, *args, **kwargs):
        if not hasattr(cls, "_cache"):
            cls._cache = dict()
        try:
            if name is not None:
                return cls._cache[name]
            else:
                raise Exception("Must provide a name parameter")
        except KeyError:
            if name is not None:
                inst = type.__call__(cls, name=name, *args, **kwargs)
                cls._cache[inst.name] = inst
                return inst
            else:
                raise KeyError("Cannot find an IonSeries for %r" % (name))


try:
    from glycopeptidepy._c.structure.fragment import IonSeriesBase as _IonSeriesBase
except ImportError:
    class _IonSeriesBase(object):
        __slots__ = ('name', 'direction', 'includes_peptide', 'mass_shift', 'regex', 'composition_shift',
                     '_hash', 'int_code')

        def __eq__(self, other):
            try:
                return self.name == other.name
            except AttributeError:
                return self.name == other

        def __ne__(self, other):
            try:
                return self.name != other.name
            except AttributeError:
                return self.name != other

        def __hash__(self):
            return self._hash


@add_metaclass(MemoizedIonSeriesMetaclass)
class IonSeries(_IonSeriesBase):
    '''An ion series type, reflecting a terminal type and a bond cleavage type.

    Attributes
    ----------
    name: str
        The name of the ion series, like 'b' or 'y'
    direction: int
        Whether the fragments from this ion series arise from the N-terminal (1),
        C-terminal (-1), or neither (0).
    includes_peptide: bool
        Whether the fragment includes the peptide backbone or not.
    mass_shift: float
        The mass delta applied to the raw components of the fragment.
    composition_shift: :class:`~.Composition`
        The elemental composition delta applied to the raw components of the fragment.
    '''

    name: str
    direction: int
    includes_peptide: bool
    mass_shift: float
    composition_shift: Composition
    int_code: int

    @classmethod
    def get(cls, name: str):
        '''Get a particular cached :class:`IonSeries` by name

        Returns
        -------
        :class:`IonSeries`
        '''
        return cls(name)

    def __init__(self, name, direction=None, includes_peptide=True, mass_shift=None, regex=None,
                 composition_shift=None, int_code=-1):
        if direction is None:
            if name in fragment_direction:
                direction = fragment_direction[name]
            else:
                direction = 0
        if mass_shift is None:
            if name in fragment_shift:
                mass_shift = fragment_shift[name]
            else:
                mass_shift = 0.
        self.name = name
        self.direction = direction
        self.includes_peptide = includes_peptide
        self.mass_shift = mass_shift
        self.regex = re.compile(regex) if regex is not None else regex
        if composition_shift is None:
            self.composition_shift = fragment_shift_composition.get(self.name, Composition())
        else:
            self.composition_shift = composition_shift
        self._hash = hash(self.name)
        self.int_code = int_code

    def __reduce__(self):
        return self.__class__, (self.name, self.direction, self.includes_peptide, self.mass_shift,
                                self.regex, self.composition_shift, self.int_code)

    def __repr__(self):
        template = ("{self.__class__.__name__}({self.name}, "
                    "direction={self.direction}, mass_shift={self.mass_shift})")
        return template.format(self=self)

    def __str__(self):
        return str(self.name)

    def is_member(self, key: str) -> bool:
        '''Test if ``key`` matches this series name.

        Parameters
        ----------
        key: str
            The name of a fragment to test for membership

        Returns
        -------
        bool
        '''
        if self.regex is None:
            return key.startswith(self.name)
        else:
            return self.regex.search(key)

    __call__ = is_member

    def __radd__(self, other):
        return other + self.mass_shift

    def __rsub__(self, other):
        return other - self.mass_shift

    def __neg__(self):
        return -self.mass_shift

    def __pos__(self):
        return self.mass_shift

    def __add__(self, other):
        return self.mass_shift + other

    def __sub__(self, other):
        return self.mass_shift - other


IonSeries.b = IonSeries("b")
IonSeries.y = IonSeries("y")
IonSeries.c = IonSeries("c")
IonSeries.z = IonSeries("z")
IonSeries.zp = IonSeries("zp")
IonSeries.z1 = IonSeries.zp

IonSeries.precursor = IonSeries("precursor")
IonSeries.oxonium_ion = IonSeries("oxonium_ion", includes_peptide=False)
IonSeries.stub_glycopeptide = IonSeries("stub_glycopeptide")
IonSeries.other = IonSeries('other')

for i, series in enumerate((IonSeries.b, IonSeries.y, IonSeries.c, IonSeries.z, IonSeries.zp, IonSeries.precursor,
                            IonSeries.oxonium_ion, IonSeries.stub_glycopeptide), 1):
    series.int_code = i



import warnings
import re
import itertools
import operator

from collections import deque

from .structure.sequence import PeptideSequence
from .structure.parser import sequence_length

_get1 = operator.itemgetter(1)
_get12 = operator.itemgetter(1, 2)


def cleave(sequence, protease, missed_cleavages=0, min_length=0, **kwargs):
    '''A reimplementation of pyteomics.parser.cleave which produces leaky cleavages
    of a peptide sequence by a regex rule. Includes the cut indices, not included in
    pyteomics.'''

    if isinstance(protease, str):
        protease = Protease(protease)
    return protease.cleave(sequence, missed_cleavages, min_length, **kwargs)


# Required information for generating peptides that can be cleaved by a given protease
enzymes = {
    "trypsin": {
        "cleavage_start": [""],
        "cleavage_end": ["K", "R"],
        "name": "trypsin"
    }
}


class Protease(object):
    def __init__(self, name, cleavage_start=None, cleavage_end=None):
        self.name = name
        self.cleavage_start = cleavage_start
        self.cleavage_end = cleavage_end
        try:
            self.regex = re.compile(enzyme_rules[name.lower()])
        except KeyError:
            self.regex = re.compile(name)

    def __eq__(self, other):
        try:
            return self.regex == other.regex
        except AttributeError:
            return False

    def __hash__(self):
        return hash(self.regex)

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return "Protease(%s, %s)" % (self.name, self.regex.pattern)

    def missed_cleavages(self, sequence):
        if isinstance(sequence, PeptideSequence):
            sequence = str(sequence)
        return len(
            list(
                filter(
                    lambda x: x.end() != len(sequence),
                    self.regex.finditer(sequence)
                )
            )
        )

    def _find_sites(self, sequence):
        return itertools.chain(
            map(lambda x: x.end(), self.regex.finditer(sequence)), [None])

    def cleave(self, sequence, missed_cleavages=0, min_length=0, max_length=100,
            semispecific=False):
        if semispecific:
            return self.cleave_semispecific(sequence, missed_cleavages, min_length, max_length)
        peptides = []
        if isinstance(sequence, PeptideSequence):
            sequence = str(sequence)
        sequence_size = sequence_length(sequence)

        all_sites = [0] + list(self._find_sites(sequence))
        for ix, i in enumerate(all_sites):
            # Update the cleavage sites to step through
            cleavage_sites = all_sites[ix:ix + missed_cleavages + 2]
            # Step through the next cleavage sites up to the missed cleavage tolerance
            for j in range(1, len(cleavage_sites)):
                # Get the sequence between this site and the jth next site
                seq = sequence[cleavage_sites[0]:cleavage_sites[j]]
                # Only deal with this if the sequence is non-empty
                if seq:
                    seq_size = sequence_length(seq)
                    if min_length is None or seq_size >= min_length:

                        # Get parent sequence coordinates
                        start_position = cleavage_sites[0]
                        end_position = cleavage_sites[j]
                        if end_position is None:
                            end_position = sequence_size
                        # Store the sequence, start position, end position, and # of missed cleavages
                        peptides.append((seq, start_position, end_position, j - 1, ))
        # Deduplicate and sort by starting postion
        return sorted(set(peptides), key=_get1)

    def cleave_semispecific(self, sequence, missed_cleavages=0, min_length=0, max_length=100):
        peptides = []
        if isinstance(sequence, PeptideSequence):
            sequence = str(sequence)

        all_sites = [0] + list(self._find_sites(sequence))
        all_sites[-1] = sequence_length(sequence)
        n_sites = len(all_sites)

        sequence_size = sequence_length(sequence)

        # See the large comment in Protease.cleave for a descrption of this first
        # pair of for-loops
        seen = dict()
        for ix, i in enumerate(all_sites):
            # Add the new site to the backtrack list
            cleavage_sites = all_sites[ix:ix + missed_cleavages + 2]

            for j in range(1, len(cleavage_sites)):
                seq = sequence[cleavage_sites[0]:cleavage_sites[j]]
                # Only deal with this if the sequence is non-empty
                if seq:
                    seq_size = sequence_length(seq)
                    if min_length is None or seq_size >= min_length:
                        # Get parent sequence coordinates
                        start_position = cleavage_sites[0]
                        end_position = cleavage_sites[j]
                        if end_position is None:
                            end_position = sequence_size
                        # Store the sequence, start position, end position, and # of missed cleavages
                        peptides.append((seq, start_position, end_position, j - 1))

            # C-terminus is enzyme-specific
            # Let the N-terminus be any position between the least recent proteolytic site
            # (cleavage_sites[0]) and the C-terminal site (i).
            i = cleavage_sites[0]
            for j in cleavage_sites[1:]:
                for k in range(i, j):
                    if k in cleavage_sites:
                        continue

                    if (k, j) in seen:
                        continue
                    seen[(k, j)] = (k, j, 'c-term')
                    seq = sequence[k:j]
                    # Only deal with this if the sequence is non-empty
                    if seq:
                        seq_size = sequence_length(seq)
                        if (min_length is None or seq_size >= min_length) and (
                                max_length is None or seq_size <= max_length):
                            # Get parent sequence coordinates
                            start_position = k
                            end_position = j
                            if end_position is None:
                                end_position = sequence_size
                            n_missed = 0
                            for site in cleavage_sites:
                                if k < site < j:
                                    n_missed += 1

                            # Store the sequence, start position, end position, and # of missed cleavages
                            peptides.append(
                                (seq, start_position, end_position, n_missed))
            # N-terminus is enzyme-specific
            # Let the C-terminus be any position between the least recent proteolytic site
            i = cleavage_sites[0]
            j = cleavage_sites[min(missed_cleavages + 2, len(cleavage_sites) - 1)]
            for k in range(j, i, -1):
                if k in cleavage_sites:
                    continue
                if (i, k) in seen:
                    continue
                seen[k, i] = (k, i, 'n-term')
                seq = sequence[i:k]
                if seq:
                    seq_size = sequence_length(seq)
                    if (min_length is None or seq_size >= min_length) and (
                            max_length is None or seq_size <= max_length):
                        # Get parent sequence coordinates
                        start_position = i
                        end_position = k
                        if end_position is None:
                            end_position = sequence_size
                        n_missed = 0
                        for site in cleavage_sites:
                            if i < site < k:
                                n_missed += 1

                        # Store the sequence, start position, end position, and # of missed cleavages
                        peptides.append(
                            (seq, start_position, end_position, n_missed))
        return sorted(set(peptides), key=_get12)

    @classmethod
    def combine(cls, *names):
        members = [Protease(name) if not isinstance(name, Protease) else name for name in names]
        patterns = merge_enzyme_rules([member.regex.pattern for member in members])
        name = ' + '.join([member.name for member in members])
        instance = cls(patterns)
        instance.name = name
        return instance

    @staticmethod
    def nonspecific_digest(sequence, min_length=0, max_length=100):
        if isinstance(sequence, PeptideSequence):
            sequence = str(sequence)
        n = len(sequence)
        out = []
        for i in range(n):
            for j in range(min_length, max_length + 1):
                if i + j > n:
                    break
                out.append((sequence[i:i + j], i, i + j, 0))
        return out


enzyme_rules = {
    'bnps-skatole': 'W',
    'caspase 1': '(?<=[FWYL]\\w[HAT])D(?=[^PEDQKR])',
    'caspase 10': '(?<=IEA)D',
    'caspase 2': '(?<=DVA)D(?=[^PEDQKR])',
    'caspase 3': '(?<=DMQ)D(?=[^PEDQKR])',
    'caspase 4': '(?<=LEV)D(?=[^PEDQKR])',
    'caspase 5': '(?<=[LW]EH)D',
    'caspase 6': '(?<=VE[HI])D(?=[^PEDQKR])',
    'caspase 7': '(?<=DEV)D(?=[^PEDQKR])',
    'caspase 8': '(?<=[IL]ET)D(?=[^PEDQKR])',
    'caspase 9': '(?<=LEH)D',
    'chymotrypsin': '(?<=[FYWL])(?!P)',
    'chymotrypsin high specificity': '([FY](?=[^P]))|(W(?=[^MP]))',
    'chymotrypsin low specificity': '([FLY](?=[^P]))|(W(?=[^MP]))|(M(?=[^PY]))|(H(?=[^DMPW]))',
    'clostripain': 'R',
    'cnbr': 'M',
    'enterokinase': '(?<=[DE]{3})K',
    'factor xa': '(?<=[AFGILTVM][DE]G)R',
    'formic acid': 'D',
    'glutamyl endopeptidase': 'E',
    'glu-c': 'E',
    'granzyme b': '(?<=IEP)D',
    'hydroxylamine': 'N(?=G)',
    'iodosobenzoic acid': 'W',
    'lys-c': "(?<=K)(?!P)",
    'ntcb': '\\w(?=C)',
    'pepsin ph1.3': '((?<=[^HKR][^P])[^R](?=[FLWY][^P]))|((?<=[^HKR][^P])[FLWY](?=\\w[^P]))',
    'pepsin ph2.0': '((?<=[^HKR][^P])[^R](?=[FL][^P]))|((?<=[^HKR][^P])[FL](?=\\w[^P]))',
    'proline endopeptidase': '(?<=[HKR])P(?=[^P])',
    'proteinase k': '[AEFILTVWY]',
    'staphylococcal peptidase i': '(?<=[^E])E',
    'thermolysin': '[^DE](?=[AFILMV])',
    'thrombin': '((?<=G)R(?=G))|((?<=[AFGILTVM][AFGILTVWA]P)R(?=[^DE][^DE]))',
    'trypsin': '([KR](?=[^P]))|((?<=W)K(?=P))|((?<=M)R(?=P))',
    # No matches ever
    'none': '^&$',
    'trypchymo': '(?<=[FYWLKR])(?!P)',
    'trypsin/p': '(?<=[KR])',
    'lys-c/p': '(?<=K)',
    'pepsina': '(?<=[FL])',
    'v8-de': '(?<=[BDEZ])(?!P)',
    'v8-e': '(?<=[EZ])(?!P)',
    '2-iodobenzoate': '(?<=W)',
    'arg-c': '(?<=R)(?!P)',
    'asp-n_ambic': '(?=[DE])',
    'asp-n': '(?=[BD])',
    'leukocyte elastase': '(?<=[ALIV])(?!P)',
    'alpha-lytic protease': '[TASV]',
}


expasy_rules = enzyme_rules


def register_enzyme(name, pattern):
    if name in enzyme_rules and enzyme_rules[name] != pattern:
        warnings.warn("Overriding {0} with rule {1} with {2}".format(name, enzyme_rules[name], pattern))
    enzyme_rules[name] = pattern


def merge_enzyme_rules(enzyme_patterns):
    rules = ["(" + pattern + ")" for pattern in enzyme_patterns]
    return "|".join(rules)


def get_enzyme(name):
    try:
        data = enzymes[name.lower()]
        return Protease(**data)
    except KeyError:
        warnings.warn("Could not identify protease {}".format(name))
        return Protease(name=name)


trypsin = Protease("trypsin")
gluc = Protease("glutamyl endopeptidase")



from collections import defaultdict


from ..structure.sequence import ProteinSequence

from .fasta import PEFFFeature, PEFFFastaHeader
from . import uniprot


class UniprotToPeffConverter(object):

    feature_to_peff_term = {
        uniprot.SignalPeptide.feature_type: ('Processed', 'PEFF:0001021'),
        uniprot.MatureProtein.feature_type: ('Processed', 'PEFF:0001020'),
        uniprot.TransitPeptide.feature_type: ('Processed', 'PEFF:0001022'),
        uniprot.Propeptide.feature_type: ('Processed', 'PEFF:0001034'),
        uniprot.SequenceVariant.feature_type: ('VariantSimple', 'PEFF:0001028'),
        uniprot.GlycosylationSite: ('ModRes', ''),
    }

    def handle_Processed(self, feature: uniprot.PeptideBase, accession):
        peff_feature = PEFFFeature(feature.start + 1, feature.end, feature.feature_type)
        return peff_feature

    def handle_VariantSimple(self, feature, accession):
        peff_feature = PEFFFeature(feature.position + 1, feature.original, feature.variation)
        return peff_feature

    def handle_GlycosylationSite(self, feature: uniprot.GlycosylationSite, accession):
        peff_feature = PEFFFeature(feature.position + 1, str(feature.glycosylation_type))
        return peff_feature

    def __call__(self, record):
        features = defaultdict(list)
        features['Prefix'] = 'sp'
        features['Tag'] = record.accessions[0]
        features['PName'] = record.recommended_name
        features['GName'] = record.gene_name
        for feature in record.features:
            try:
                key = self.feature_to_peff_term[feature.feature_type]
                handler = getattr(self, 'handle_' + key[0], None)
                if handler is None:
                    continue
                features[key[0]].append(handler(feature, key[1]))
            except KeyError:
                continue
        header = PEFFFastaHeader(dict(features))
        return ProteinSequence(header, record.sequence)



from .term import Term


class PEFFTerm(Term):
    __slots__ = ()


peff_cv_terms = []

# [[[cog
# import cog
# from glycopeptidepy.io.cv.term import render_list
# render_list('PEFF CV Term', list_name="peff_cv_terms", term_cls_name="PEFFTerm", writer=cog.out)
# ]]]
peff_cv_terms = [
    PEFFTerm(u'PEFF file sequence entry term', u'PEFF:0000003',
             u'CV term that may appear in a description line of a PEFF file individual sequence entry.',       'PEFF CV Term', [u'PEFF CV term']),
    PEFFTerm(u'PEFF file header section term', u'PEFF:0000002',
             u'CV term that may appear in a PEFF file header section.',       'PEFF CV Term', [u'PEFF CV term']),
    PEFFTerm(u'NcbiTaxId', u'PEFF:0001003', u'PEFF keyword for the NCBI taxonomy identifier.',
             'PEFF CV Term', [u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'TaxName', u'PEFF:0001004', u'PEFF keyword for the taxonomy name (latin or common name).',
             'PEFF CV Term', [u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'Length', u'PEFF:0001006', u'PEFF keyword for the sequence length.',
             'PEFF CV Term', [u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'SV', u'PEFF:0001007', u'PEFF keyword for the sequence version.',
             'PEFF CV Term', [u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'EV', u'PEFF:0001008', u'PEFF keyword for the entry version.',
             'PEFF CV Term', [u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'PE', u'PEFF:0001009', u'PEFF keyword for the Protein Evidence; A UniProtKB code 1-5.',
             'PEFF CV Term', [u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'DbUniqueId', u'PEFF:0001001', u'OBSOLETE Sequence database unique identifier.',
             'PEFF CV Term', [u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'PName', u'PEFF:0001002', u'PEFF keyword for the protein full name.',
             'PEFF CV Term', [u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'XRef', u'PEFF:0001019', u'PEFF keyword for the cross-reference to an external resource.',
             'PEFF CV Term', [u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'GO', u'PEFF:0001018', u'PEFF keyword for the Gene Ontology code.',
             'PEFF CV Term', [u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'ModRes', u'PEFF:0001013', u'PEFF keyword for the modified residue without aPSI-MOD or UniMod identifier.',
             'PEFF CV Term', [u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'ModResPsi', u'PEFF:0001012', u'PEFF keyword for the modified residue with PSI-MOD identifier.',
             'PEFF CV Term', [u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'Variant', u'PEFF:0001011', u'OBSOLETE Sequence variation (substitution, insertion, deletion).',
             'PEFF CV Term', [u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'Processed', u'PEFF:0001010', u'PEFF keyword for information on how the full length original protein sequence can be processed into shorter components such as signal peptides and chains.',
             'PEFF CV Term', [u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'KW', u'PEFF:0001017', u'PEFF keyword for the entry associated keyword(s).',
             'PEFF CV Term', [u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'CC', u'PEFF:0001016', u'PEFF keyword for the entry associated comment.',
             'PEFF CV Term', [u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'SeqStatus', u'PEFF:0001015', u'PEFF keyword for the sequence status. Complete or Fragment.',
             'PEFF CV Term', [u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'AltAC', u'PEFF:0001014', u'PEFF keyword for the Alternative Accession Code.',
             'PEFF CV Term', [u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'ID', u'PEFF:0001026', u'PEFF keyword for the UniProtKB specific Protein identifier ID; a UniProtKB term.',
             'PEFF CV Term', [u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'ModResUnimod', u'PEFF:0001027', u'PEFF keyword for the modified residue with UniMod identifier.',
             'PEFF CV Term', [u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'Crc64', u'PEFF:0001024', u'PEFF keyword for the Sequence checksum in crc64.',
             'PEFF CV Term', [u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'Domain', u'PEFF:0001025', u'PEFF keyword for the sequence range of a domain.',
             'PEFF CV Term', [u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'Conflict', u'PEFF:0001023', u'PEFF keyword for the sequence conflict; a UniProtKB term.',
             'PEFF CV Term', [u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'VariantSimple', u'PEFF:0001028', u'PEFF keyword for the simple sequence variation of a single amino acid change. A change to a stop codon is permitted with a * symbol. More complex variations must be encoded with the VariantComplex term.',
             'PEFF CV Term', [u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'VariantComplex', u'PEFF:0001029', u'PEFF keyword for a sequence variation that is more complex than a single amino acid change or change to a stop codon.',
             'PEFF CV Term', [u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'GName', u'PEFF:0001005', u'PEFF keyword for the gene name.',
             'PEFF CV Term', [u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'DisulfideBond', u'PEFF:0001031', u'PEFF keyword for the disulfide bonds in this protein, constructed as a sets of annotation identifiers of two half-cystine modifications.',
             'PEFF CV Term', [u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'Proteoform', u'PEFF:0001030', u'PEFF keyword for the proteoforms of this protein, constructed as a set of annotation identifiers.',
             'PEFF CV Term', [u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'Comment', u'PEFF:0001033', u'PEFF keyword for the individual protein entry comment. It is discouraged to put parsable information here. This is only for free-text commentary.',
             'PEFF CV Term', [u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'PEFF molecule processing keyword', u'PEFF:0001032', u'PEFF keyword describing the type of processing event being described.',
             'PEFF CV Term', [u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'ProteoformDb', u'PEFF:0000022', u"PEFF keyword that when set to 'true' indicates that the database contains complete proteoforms.",
             'PEFF CV Term', [u'PEFF file header section term', u'PEFF CV term']),
    PEFFTerm(u'SequenceType', u'PEFF:0000017', u'PEFF keyword for the molecular type of the sequences.',
             'PEFF CV Term', [u'PEFF file header section term', u'PEFF CV term']),
    PEFFTerm(u'GeneralComment', u'PEFF:0000021', u'PEFF keyword for a general comment.',
             'PEFF CV Term', [u'PEFF file header section term', u'PEFF CV term']),
    PEFFTerm(u'DatabaseDescription', u'PEFF:0000020', u'PEFF keyword for the short description of the PEFF file.',
             'PEFF CV Term', [u'PEFF file header section term', u'PEFF CV term']),
    PEFFTerm(u'OptionalTagDef', u'PEFF:0000023', u'PEFF keyword for the short tag (abbreviation) and longer definition used to annotate a sequence annotation (such as variant or modification) in the OptionalTag location.',
             'PEFF CV Term', [u'PEFF file header section term', u'PEFF CV term']),
    PEFFTerm(u'HasAnnotationIdentifiers', u'PEFF:0000024', u"PEFF keyword that when set to 'true' indicates that entries in the database have identifiers for each annotation.",
             'PEFF CV Term', [u'PEFF file header section term', u'PEFF CV term']),
    PEFFTerm(u'Prefix', u'PEFF:0000009', u'PEFF keyword for the sequence database prefix.',
             'PEFF CV Term', [u'PEFF file header section term', u'PEFF CV term']),
    PEFFTerm(u'DbName', u'PEFF:0000008', u'PEFF keyword for the sequence database name.',
             'PEFF CV Term', [u'PEFF file header section term', u'PEFF CV term']),
    PEFFTerm(u'NumberOfEntries', u'PEFF:0000015', u'PEFF keyword for the sumber of sequence entries in the database.',
             'PEFF CV Term', [u'PEFF file header section term', u'PEFF CV term']),
    PEFFTerm(u'DbDescription', u'PEFF:0000010', u'PEFF keyword for the sequence database short description.',
             'PEFF CV Term', [u'PEFF file header section term', u'PEFF CV term']),
    PEFFTerm(u'Decoy', u'PEFF:0000011', u'PEFF keyword for the specifying whether the sequence database is a decoy database.',
             'PEFF CV Term', [u'PEFF file header section term', u'PEFF CV term']),
    PEFFTerm(u'DbVersion', u'PEFF:0000013', u'PEFF keyword for the database version (release date) according to database provider.',
             'PEFF CV Term', [u'PEFF file header section term', u'PEFF CV term']),
    PEFFTerm(u'DbDate', u'PEFF:0000014', u'OBSOLETE PEFF keyword for the database date (release or file date of the source) according to database provider.',
             'PEFF CV Term', [u'PEFF file header section term', u'PEFF CV term']),
    PEFFTerm(u'SpecificValue', u'PEFF:0000019', u'PEFF keyword for the specific values for a custom key.',
             'PEFF CV Term', [u'PEFF file header section term', u'PEFF CV term']),
    PEFFTerm(u'SpecificKey', u'PEFF:0000018', u'PEFF keyword for database specific keywords not included in the current controlled vocabulary.',
             'PEFF CV Term', [u'PEFF file header section term', u'PEFF CV term']),
    PEFFTerm(u'DbSource', u'PEFF:0000012', u'PEFF keyword for the source of the database file.',
             'PEFF CV Term', [u'PEFF file header section term', u'PEFF CV term']),
    PEFFTerm(u'Conversion', u'PEFF:0000016', u'PEFF keyword for the description of the conversion from original format to this current one.',
             'PEFF CV Term', [u'PEFF file header section term', u'PEFF CV term']),
    PEFFTerm(u'transit peptide', u'PEFF:0001022', u'Short peptide present at the N-terminus of a newly synthesized protein that helps the protein through the membrane of its destination organelle.',
             'PEFF CV Term', [u'PEFF molecule processing keyword', u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'mature protein', u'PEFF:0001020', u'Portion of a newly synthesized protein that contributes to a final structure after other components such as signal peptides are removed.',
             'PEFF CV Term', [u'PEFF molecule processing keyword', u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'signal peptide', u'PEFF:0001021', u'Short peptide present at the N-terminus of a newly synthesized protein that is cleaved off and is not part of the final mature protein.',
             'PEFF CV Term', [u'PEFF molecule processing keyword', u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'initiator methionine', u'PEFF:0001035', u'N-terminal methionine residue of a protein that can be co-translationally cleaved.',
             'PEFF CV Term', [u'PEFF molecule processing keyword', u'PEFF file sequence entry term', u'PEFF CV term']),
    PEFFTerm(u'propeptide', u'PEFF:0001034', u'Short peptide that is cleaved off a newly synthesized protein and generally immediately degraded in the process of protein maturation, and is not a signal peptide or transit peptide.',
             'PEFF CV Term', [u'PEFF molecule processing keyword', u'PEFF file sequence entry term', u'PEFF CV term']),
]
# [[[end]]]

peff_cv_terms_by_name = {c.name: c for c in peff_cv_terms}


def peff_cv_term(name, strict=False):
    try:
        return peff_cv_terms_by_name[name]
    except KeyError:
        if not strict:
            return PEFFTerm(name, name, name, name, [name])
        else:
            raise



from collections import defaultdict

try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping

import itertools

try:
    range = xrange
except Exception:
    pass


def _identity(i):
    return i


def groupby(ungrouped_list, key_fn=_identity, transform_fn=_identity):
    groups = defaultdict(list)
    for item in ungrouped_list:
        key_value = key_fn(item)
        groups[key_value].append(transform_fn(item))
    return groups


def descending_combination_counter(counter):
    keys = counter.keys()
    count_ranges = map(lambda x: range(x + 1), counter.values())
    for combination in itertools.product(*count_ranges):
        yield dict(zip(keys, combination))


class decoratordict(dict):
    def __call__(self, key):
        def wrapper(f):
            self[key] = f
            return f
        return wrapper


class _AccumulatorBag(Mapping):
    def __init__(self, source=None):
        self.store = defaultdict(int)
        if source is not None:
            self.store.update(source)

    def __getitem__(self, i):
        return self.store[i]

    def __setitem__(self, i, v):
        self.store[i] = v

    def __add__(self, other):
        new = _AccumulatorBag(self)
        for key, value in other.items():
            new[key] += value
        return new

    def __iadd__(self, other):
        for key, value in other.items():
            self[key] += value
        return self

    def __iter__(self):
        return iter(self.store)

    def __len__(self):
        return len(self.store)

    def __contains__(self, key):
        return key in self.store

    def keys(self):
        return self.store.keys()

    def values(self):
        return self.store.values()

    def items(self):
        return self.store.items()

    def __repr__(self):
        return "_AccumulatorBag(%r)" % dict(self.store)


try:
    _descending_combination_counter = descending_combination_counter
    from glycopeptidepy._c.collectiontools import descending_combination_counter
except ImportError:
    has_c = False

try:
    from glycopeptidepy._c.count_table import CountTable

    del _AccumulatorBag

    class _AccumulatorBag(CountTable):
        pass
except ImportError:
    has_c = False


__all__ = [
    "descending_combination_counter", "_AccumulatorBag"
]
